@@ -29,6 +29,7 @@
 #include <errno.h>
 #include "config.h"
 #include <assert.h>
+#include <libgen.h>
 
 #if defined(__APPLE__) && defined(__MACH__)
 
@@ -39,21 +40,20 @@
 #endif
 
 #include "xmalloc.h"
-#include "scorer.h"
 #include "filtration.h"
 #include "vector.h"
 #include "timing.h"
 #include "inline_qsort.h"
 #include "loading.h"
+#include "do_with_main_loop.h"
 
 static GtkWindow *top_window;
 static GtkEntry *name_entry;
 static GtkTreeView *tree_view;
 static GtkListStore *list_store;
 
-struct vector filtered = {.eltsize = sizeof(struct filter_result)};
-
 static char *init_filter;
+static gboolean show_version;
 static gboolean multiselect;
 static char *project_type;
 static char *project_dir;
@@ -63,6 +63,19 @@ static gboolean disable_hg;
 
 static
 gboolean program_exited;
+static
+gboolean gpicker_loading_completed;
+static
+struct do_with_main_loop_process *loading_process;
+
+GPid child_pid;
+
+static
+void kill_child_pid(void)
+{
+	if (child_pid)
+		kill(child_pid, SIGINT);
+}
 
 static void filter_tree_view_tail();
 
@@ -139,10 +152,6 @@ void cell_data_func(GtkTreeViewColumn *col,
 		const char *pattern = applied_pattern;
 		int patlen = strlen(pattern);
 		unsigned match[patlen];
-		void *filter;
-		filter_func filter_func;
-		filter_destructor destructor = 0;
-		struct filter_result result;
 		int passes;
 		int i;
 
@@ -150,18 +159,19 @@ void cell_data_func(GtkTreeViewColumn *col,
 			     "text", text,
 			     NULL);
 
-		filter = prepare_filter(pattern, &filter_func, &destructor);
-		passes = filter_func(files + index, filter, &result, match);
-		if (destructor)
-			destructor(filter);
+		passes = obtain_match(pattern, index, match);
 		if (!passes)
 			return;
 
 		PangoAttrList *list = pango_attr_list_new();
 		for (i=0;i<patlen;i++) {
+			unsigned match_pos = match[i];
+			if (match_pos == SCORER_MATCH_NONE)
+				continue;
 			PangoAttribute *attr = pango_attr_weight_new(PANGO_WEIGHT_BOLD);
-			attr->start_index = match[i];
-			attr->end_index = match[i]+1;
+			attr->start_index = match_pos;
+			while (utf8_continuation_p(text[++match_pos]));
+			attr->end_index = match_pos;
 			pango_attr_list_insert(list, attr);
 		}
 
@@ -194,9 +204,10 @@ static
 void exit_program(void)
 {
 	program_exited = TRUE;
-	if (!gpicker_loading_completed)
+	if (!gpicker_loading_completed) {
 		read_filenames_abort();
-	else
+		do_with_main_loop_quit(loading_process);
+	} else
 		gtk_main_quit();
 }
 
@@ -226,8 +237,6 @@ void print_selection(void)
 {
 	GtkTreeSelection *sel = gtk_tree_view_get_selection(tree_view);
 	GList *list;
-	GtkTreeIter iter;
-	gint idx;
 
 	list = gtk_tree_selection_get_selected_rows(sel, 0);
 	if (!list)
@@ -325,12 +334,13 @@ int my_popen(char *string, GPid *child_pid)
 }
 
 static
-void setup_filenames(void)
+gpointer setup_filenames_core(gpointer _dummy,
+			      struct do_with_main_loop_process *p)
 {
-	int pipe;
-	GPid pid = 0;
+	int pipe = -1;
 
 	if (project_type && !strcmp(project_type, "mlocate")) {
+		do_with_main_loop_init_complete(p);
 		int fd = open(project_dir, O_RDONLY);
 		if (fd < 0) {
 			perror("setup_filenames:open");
@@ -338,7 +348,7 @@ void setup_filenames(void)
 		}
 		read_filenames_from_mlocate_db(fd);
 		close(fd);
-		return;
+		return 0;
 	}
 
 	if (read_stdin)
@@ -347,30 +357,44 @@ void setup_filenames(void)
 		char *find_command = getenv("GPICKER_FIND");
 		if (!find_command)
 			find_command = FIND_INVOCATION;
-		pipe = my_popen(find_command, &pid);
+		pipe = my_popen(find_command, &child_pid);
 	} else if (!strcmp(project_type, "git"))
-		pipe = my_popen("git ls-files --exclude-standard -c -o -z .", &pid);
+		pipe = my_popen("git ls-files --exclude-standard -c -o -z .", &child_pid);
 	else if (!strcmp(project_type, "hg"))
-		pipe = my_popen("hg locate -0 --include .", &pid);
+		pipe = my_popen("hg locate -0 --include .", &child_pid);
 	else if (!strcmp(project_type, "bzr"))
-		pipe = my_popen("bzr ls -R --versioned --unknown --null", &pid);
+		pipe = my_popen("bzr ls -R --versioned --unknown --null", &child_pid);
+
+	do_with_main_loop_init_complete(p);
 
 	if (pipe < 0) {
 		perror("failed to spawn find");
 		exit(1);
 	}
 
+	read_filenames(pipe);
+
+	if (child_pid) {
+		kill(child_pid, SIGINT);
+		child_pid = 0;
+	}
+	close(pipe);
+
+	return 0;
+}
+
+static
+void setup_filenames(void)
+{
 	g_timeout_add(250, async_load_callback, 0);
 	gtk_widget_set_sensitive(GTK_WIDGET(tree_view), FALSE);
 
-	read_filenames_with_main_loop(pipe);
+	do_with_main_loop((do_with_main_loop_fn)setup_filenames_core, 0, &loading_process);
+	gpicker_loading_completed = TRUE;
+	do_with_main_loop_close(loading_process);
 
 	gtk_widget_set_sensitive(GTK_WIDGET(tree_view), TRUE);
 	set_window_title();
-
-	if (pid)
-		kill(pid, SIGINT);
-	close(pipe);
 }
 
 static
@@ -437,14 +461,16 @@ void setup_signals(void)
 
 static
 GOptionEntry entries[] = {
+	{"version", 0, 0, G_OPTION_ARG_NONE, &show_version, "show version", 0},
 	{"project-type", 't', 0, G_OPTION_ARG_STRING, &project_type, "respect ignored files for given kind of VCS (default, git, bzr, hg, guess)", 0},
 	{"disable-bzr", 0, 0, G_OPTION_ARG_NONE, &disable_bzr, "disable autodetection of Bazaar project type", 0},
 	{"disable-hg", 0, 0, G_OPTION_ARG_NONE, &disable_hg, "disable autodetection of Mercurial project type", 0},
-	{"name-separator", 0, 0, G_OPTION_ARG_STRING, &name_separator, "separator of filenames from stdin (\\0 is default)", 0},
-	{"dir-separator", 0, 0, G_OPTION_ARG_STRING, &dir_separator, "separator of directory names from stdin (/ is default)", 0},
+	{"name-separator", 'n', 0, G_OPTION_ARG_STRING, &name_separator, "separator of filenames from stdin (\\0 is default)", 0},
+	{"dir-separator", 'd', 0, G_OPTION_ARG_STRING, &dir_separator, "separator of directory names from stdin (/ is default)", 0},
 	{"eat-prefix", 0, 0, G_OPTION_ARG_STRING, &eat_prefix, "eat this prefix from names (./ is default)", 0},
-	{"multiselect", 0, 0, G_OPTION_ARG_NONE, &multiselect, "enable multiselect", 0},
-	{"init-filter", 0, 0, G_OPTION_ARG_STRING, &init_filter, "initial filter value", 0},
+	{"multiselect", 'm', 0, G_OPTION_ARG_NONE, &multiselect, "enable multiselect", 0},
+	{"init-filter", 'i', 0, G_OPTION_ARG_STRING, &init_filter, "initial filter value", 0},
+	{"load-stdin-too", 0, 0, G_OPTION_ARG_NONE, &gpicker_load_stdin_too, "read additional filenames from stdin", 0},
 	{0}
 };
 
@@ -547,6 +573,10 @@ void parse_options(int argc, char **argv)
 		fprintf(stderr, "option parsing failed: %s\n", error->message);
 		exit(1);
 	}
+	if (show_version) {
+		fprintf(stdout, "%s %s\n", PACKAGE_NAME, VERSION);
+		exit(0);
+	}
 	if (argc < 2) {
 		fprintf(stderr, "I need a project path\n");
 		fputs(g_option_context_get_help(context, TRUE, NULL), stderr);
@@ -632,14 +662,23 @@ void build_ui()
 	gtk_widget_show_all(GTK_WIDGET(top_window));
 }
 
+extern
+int simple_main(int, char **);
+
 int main(int argc, char **argv)
 {
+	char *gpicker = basename(xstrdup(argv[0]));
+	if (!strcmp(gpicker, "gpicker-simple"))
+		return simple_main(argc, argv);
+
 	timing_t tstart = start_timing();
 
 	g_thread_init(0);
 	parse_options(argc, argv);
 	gtk_init(0, 0);
 
+	atexit(kill_child_pid);
+
 	finish_timing(tstart, "gtk initialization");
 	tstart = start_timing();
 
@@ -681,6 +720,8 @@ int main(int argc, char **argv)
 
 	gdk_window_set_cursor(GTK_WIDGET(top_window)->window, 0);
 
+	gtk_widget_grab_focus(GTK_WIDGET(name_entry));
+
 	finish_timing(tstart, "setup_data");
 
 	if (!program_exited)
